import pytest
import json
from time import time

from a3 import Embedding, BinaryHeap, find_similar


@pytest.fixture(scope="session", autouse=True)
def word2embedding():
    return json.load(open("food.json", "r", encoding="utf-8"))


def test_heap_int():
    heap = BinaryHeap()
    heap.push(11)
    heap.push(1)
    heap.push(4)
    heap.push(5)
    heap.push(3)
    assert str(heap) == "(11 (5 1 3) 4)"
    assert heap.pop() == 11
    assert str(heap) == "(5 (3 1) 4)"
    
    
def test_heap_pairs():
    pairs = [("a", 1), ("z", 2), ("d", 3), ("b", 4), ("w", 5)]
    heap1 = BinaryHeap(key=lambda x: x[0])
    heap2 = BinaryHeap(key=lambda x: x[1])
    for pair in pairs:
        heap1.push(pair)
        heap2.push(pair)
    assert str(heap1) == "(('z', 2) (('w', 5) ('a', 1) ('b', 4)) ('d', 3))"
    assert str(heap2) == "(('w', 5) (('b', 4) ('a', 1) ('d', 3)) ('z', 2))"


def test_pop_at_root():
    heap = BinaryHeap()
    heap.push("arancina")
    assert heap.pop() == "arancina"
    assert len(heap.H) == 0


def test_find_similar(word2embedding):
    expected = {
        "pasta": "[lasagna, ravioli, soup]",
        "strawberry": "[blueberry, peach, watermelon]",
        "borsch": "[bliny, squashberry, jostaberry]"
    }
    for target_word in expected.keys():
        target_embedding = Embedding(target_word, word2embedding[target_word])
        embeddings = list()
        for word in word2embedding:
            embeddings.append(Embedding(word, word2embedding[word]))
        assert str(find_similar(target_embedding, embeddings))\
            == expected[target_word]


def test_target_in_first_k(word2embedding):
    target_word = "food"
    target_embedding = Embedding(target_word, word2embedding[target_word])
    embeddings = list()
    for word in word2embedding:
        embeddings.append(Embedding(word, word2embedding[word]))
    top = find_similar(target_embedding, embeddings, k=500)
    assert len(top) == 500
    assert str(top[0]) != "food"


def find_similar_bf(target_embedding, embeddings, k=3):
    """
    Brute force implementation of the find_similar function.
    """
    similar = []
    for _ in range(k):
        mindist, minj = float('inf'), None
        for j, emb in enumerate(embeddings):
            d = target_embedding.distance(emb)
            if emb == target_embedding or j in similar: continue
            if d < mindist:
                mindist, minj = d, j
        if minj: similar.append(minj)
    return [embeddings[j] for j in similar]


def test_performance(word2embedding):
    target_word = "pizza"
    target_embedding = Embedding(target_word, word2embedding[target_word])
    embeddings = list()
    for word in word2embedding:
        embeddings.append(Embedding(word, word2embedding[word]))
    
    start = time()
    _ = find_similar_bf(target_embedding, embeddings, k=100)
    bf_time = time() - start

    start = time()
    _ = find_similar(target_embedding, embeddings, k=100)
    our_time = time() - start

    assert our_time < bf_time
